"""Inquisitor settings management."""

import logging
from functools import lru_cache
from typing import cast

from cryptography.hazmat.primitives import hashes
from fastapi import Depends
from frgal import make_credentials
from grill import Logger, LoggerClient
from pydantic import BaseModel
from redis.asyncio import Redis
from regal import ContactClient, DomainClient, RegistrarClient
from setapp import BaseApplication, BaseSettings, SettingsConfigDict
from verifier import VerificationClient

from .constants import LOGGER_SERVICE, LoggerResult
from .utils import load_instance

_LOGGER = logging.getLogger(__name__)


class OIDCSettings(BaseModel):
    """Inquisitor OIDC settings."""

    issuer: str
    client_id: str
    client_secret: str


class Settings(BaseSettings):
    """Actual settings."""

    model_config = SettingsConfigDict(
        config_environ="INQUISITOR_CONFIG",
        config_files=("~/.fred/inquisitor.conf", "/etc/fred/inquisitor.conf"),
        env_prefix="inquisitor_",
    )

    secret_key: str
    allowed_origins: list[str] = []
    access_token_expires: int = 60 * 60
    registry_netloc: str
    registry_ssl_cert: str | None = None
    logger_netloc: str
    logger_ssl_cert: str | None = None
    verification_netloc: str
    verification_ssl_cert: str | None = None
    redis_cache_cls: str
    redis_cache_options: dict = {}
    redis_cache_key_prefix: str = ""
    oidc: OIDCSettings
    max_letters: int = 2

    def __hash__(self) -> int:
        # Make settings hashable to be used as argument in cached function
        # See https://github.com/pydantic/pydantic/issues/1303
        return hash((type(self),) + tuple(self.model_dump_json()))

    @property
    def secret_key_sha256(self) -> bytes:
        """Generate SHA256 hash from secret key."""
        digest = hashes.Hash(hashes.SHA256())
        digest.update(self.secret_key.encode())
        return digest.finalize()


class Application(BaseApplication[Settings]):
    """Represent application settings and basic setup."""


@lru_cache
def get_app() -> Application:
    """Get application instance."""
    APP = Application(Settings)
    APP.load()
    return APP


@lru_cache
def get_settings() -> Settings:
    """Get settings instance."""
    app = get_app()
    return app.settings


@lru_cache
def get_contact_client(settings: Settings = Depends(get_settings)) -> ContactClient:
    """Get contact client."""
    return ContactClient(
        settings.registry_netloc,
        credentials=make_credentials(settings.registry_ssl_cert),
    )


@lru_cache
def get_domain_client(settings: Settings = Depends(get_settings)) -> DomainClient:
    """Get domain client."""
    return DomainClient(
        settings.registry_netloc,
        credentials=make_credentials(settings.registry_ssl_cert),
    )


@lru_cache
def get_registrar_client(settings: Settings = Depends(get_settings)) -> RegistrarClient:
    """Get registrar client."""
    return RegistrarClient(
        settings.registry_netloc,
        credentials=make_credentials(settings.registry_ssl_cert),
    )


@lru_cache
def get_verification_client(settings: Settings = Depends(get_settings)) -> VerificationClient:
    """Get verification client."""
    return VerificationClient(
        settings.verification_netloc,
        credentials=make_credentials(settings.verification_ssl_cert),
    )


@lru_cache
def get_logger_client(settings: Settings = Depends(get_settings)) -> LoggerClient:
    """Get logger client."""
    return LoggerClient(
        settings.logger_netloc,
        credentials=make_credentials(settings.logger_ssl_cert),
    )


@lru_cache
def get_logger(logger_client: LoggerClient = Depends(get_logger_client)) -> Logger:
    """Get logger."""
    return Logger(logger_client, service=LOGGER_SERVICE, default_result=LoggerResult.ERROR)


@lru_cache
def get_cache(settings: Settings = Depends(get_settings)) -> Redis:
    """Get cache instance."""
    return cast(Redis, load_instance(settings.redis_cache_cls, **settings.redis_cache_options))
